//****************************************************************************
//       Copyright (C) 2011 Hartmut Monien <monien@th.physik.uni-bonn.de>
//
//  Distributed under the terms of the GNU General Public License (GPL)
//
//    This code is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//    General Public License for more details.
//
//  The full text of the GPL is available at:
//
//                  http://www.gnu.org/licenses/
//****************************************************************************

#include "sl2z.hpp"

const SL2Z SL2Z::S(0, -1, 1, 0);
const SL2Z SL2Z::U(0,  1,-1, 0);
const SL2Z SL2Z::R(1, -1, 0, 1);
const SL2Z SL2Z::T(1,  1, 0, 1);
const SL2Z SL2Z::E(1,  0, 0, 1);
const SL2Z SL2Z::I(-1, 0, 0, -1);

